/*
 * This file contains the interface for a socket wrapper to provide some basic
 * TLS client support using the openssl library.  Openssl is a well-respected
 * product (despite one very famous bug on the server side), but they didn't
 * write this wapper, and it may have some interesting holes due to poor use of
 * the library.  Also note that it is client-side only, and tends to use
 * defaults, rather than exposing much of the wide configurability of the
 * underlying openssl libary.  Perhaps some of these things will be remedied
 * one day, but I want to get something working before the very last plain-text
 * server disappears from the network, and I can't assign any more programs
 * :-).
 *
 * The object loads a .pem file representing the root authority from either a
 * default location, or one specified with the authority_file() method (which
 * is generally safer).  When the authority_file method is not used, the
 * default behavior is to search through several locations controled by the
 * program environment.  The wrapper object created in the usual way will
 * assume the underlying socket is properly connected and attempt and TLS/SSL
 * handshake with the remote server.  If the second constructor parameter is
 * given as false, the object is created without a secure connection, and
 * operates as the underlying object until the start_tls() call is made, which
 * completes the connection.
 */

/*
 * This file is part of the Cleansocks library.  It is released under the
 * Lesser GNU GPL.  See the accompanying files COPYING and COPYING.LESSER.
 */
#ifndef _cleantslsc_h_
#define _cleantslsc_h_

#include <stdio.h>

#include <iostream>
#include <string>

// The only references to stuff from the openssl headers are two pointers in
// in the client_tls_socket class.  I want to avoid the need for the headers
// in the client distribution, at least for Windows.  
#ifdef BUILDLIB
#include <openssl/ssl.h>
#include <openssl/err.h>
#else
typedef void SSL_CTX;
typedef void SSL;
#endif

#include "cleansri.h"
#include "cleansocks.h"
#include "cleanflags.h"

namespace cleansocks {

/* Flag values to send to newly-created secure sockets. */
typedef clean_flags_tmpl<2> cleantlsc_flags;

// Don't check the server certificate.  If this is set, the following _OK
// flags are ignored.
static const cleantlsc_flags NO_CERT_CHECK(1);

// Self-signed cert is okay.
static const cleantlsc_flags SELF_OK(2);

// An expired certificate will be accepted anyway.
static const cleantlsc_flags EXPIRED_OK(4);

// A server may have no certificate at all.
static const cleantlsc_flags NO_CERT_OK(8);

// Suppress the SNI extension, even though a host name was provided.
static const cleantlsc_flags NO_SNI(16);

// Suppress certificate subject host name check, even though a host name was provided.
static const cleantlsc_flags NO_HOST_CHK(32);

class client_tls_socket: public sr_obj {
public:
	static void authority_file(std::string path) {
		sm_authority_path = path;
	}
	client_tls_socket(): m_base(nullptr), m_flags(0), m_certhost(""), m_ssl(nullptr) { }
	client_tls_socket(sr_obj &base, cleantlsc_flags flags, std::string certhost = "",
			  bool secure_now = true):
			m_base(&base), m_flags(flags), m_certhost(certhost), m_ssl(nullptr)
	{
		allocate_context();
		if(secure_now) start_tls();
	}
	void operator()(sr_obj &base, cleantlsc_flags flags, std::string certhost = "",
			bool secure_now = true) {
		reinitchk();
		m_base = &base;
		m_flags = flags;
		m_certhost = certhost;
		allocate_context();
		if(secure_now) start_tls();
	}
	client_tls_socket(sr_obj &base, cleantlsc_flags flags, bool secure_now):
			m_base(&base), m_flags(flags), m_certhost(""), m_ssl(nullptr)
	{
		allocate_context();
		if(secure_now) start_tls();
	}
	void operator()(sr_obj &base, cleantlsc_flags flags, bool secure_now)
	{
		reinitchk();
		m_base = &base;
		m_flags = flags;
		allocate_context();
		if(secure_now) start_tls();
	}
	client_tls_socket(sr_obj &base, std::string certhost, bool secure_now = true):
			m_base(&base), m_flags(0), m_certhost(certhost), m_ssl(nullptr)
	{
		allocate_context();
		if(secure_now) start_tls();
	}
	void operator()(sr_obj &base, std::string certhost, bool secure_now = true)
	{
		reinitchk();
		m_base = &base;
		m_certhost = certhost;
		allocate_context();
		if(secure_now) start_tls();
	}
	client_tls_socket(sr_obj &base, bool secure_now):
			m_base(&base), m_flags(0), m_certhost(""), m_ssl(nullptr)
	{
		allocate_context();
		if(secure_now) start_tls();
	}
	void operator()(sr_obj &base, bool secure_now)
	{
		reinitchk();
		m_base = &base;
		allocate_context();
		if(secure_now) start_tls();
	}
	void start_tls() { start_tls_private(nullptr); }
	void start_tls_share_session(client_tls_socket &sock) {
		start_tls_private(&sock);
	}
	bool tls_active() { return m_ssl != nullptr; }
	virtual int send(const void *buf, int size);
	virtual int recv(void *buf, int size);
	virtual void close() {
		stop_tls();
		cleansocks::close(*m_base);
	}
	void stop_tls();
protected:
	void reinitchk() {
		if(m_base != nullptr)
			throw cleansocks_reinit_error("client_tls_socket reinitialized.");
	}
	
	void allocate_context();

	void start_tls_private(client_tls_socket *sock);

	sr_obj *m_base;			// Underlying socket.
	cleantlsc_flags m_flags;	// Behavior flags from construction.

	// The host name for cert checking.
	std::string m_certhost;
	
	// TLS stuff.
	static SSL_CTX *sm_ctx;		// SSL context, shared over cleantlsc.
	SSL *m_ssl;			// SSL endpoint.

	client_tls_socket(const client_tls_socket&) = delete;
	client_tls_socket& operator=(const client_tls_socket&) = delete;

private:
	static std::string sm_authority_path;	// User set authority path
	static std::string sm_found_authority;	// Authority found on pack init
	static std::string find_authority();	// This sets the above.
};

class socket_tls_error: public socket_error {
public:
	// Print the full error stack.
	static void error_stack();
protected:
	socket_tls_error(std::string msg, bool get_ssl_err = true): 
		socket_error(get_ssl_err ? get_err(msg) : msg) { }
private:
	static std::string get_err(std::string msg);
};

class socket_tls_context_create: public socket_tls_error {
public:
	socket_tls_context_create(): 
		socket_tls_error("Error creating SSL context") { }
};

class socket_tls_set_ca: public socket_tls_error {
public:
	socket_tls_set_ca(std::string ca): 
		socket_tls_error("Error setting CA ["+ca+"]") { }
};

class socket_tls_ssl_create: public socket_tls_error {
public:
	socket_tls_ssl_create(): 
		socket_tls_error("Error creating SSL object") { }
};

class socket_tls_creating_bio: public socket_tls_error {
public:
	socket_tls_creating_bio(): 
		socket_tls_error("Error creating TLS BIO object") { }
};

class socket_tls_handshake: public socket_tls_error {
public:
	socket_tls_handshake(): 
		socket_tls_error("TLS handshake failed") { }
};

class socket_tls_ca_path: public socket_tls_error {
public:
	socket_tls_ca_path(): 
		socket_tls_error("No CA path found", false) { }
};

class socket_tls_send: public socket_tls_error {
public:
	socket_tls_send(): 
		socket_tls_error("TLS write failed") { }
protected:
	socket_tls_send(std::string msg): 
		socket_tls_error(msg) { }
};

class socket_tls_send_shut: public socket_tls_send {
public:
	socket_tls_send_shut(): 
		socket_tls_send("TLS write failed due to shutdown") { }
};

class socket_tls_recv: public socket_tls_error {
public:
	socket_tls_recv(): 
		socket_tls_error("TLS read failed") { }
};

class socket_tls_bad_cert: public socket_tls_error {
public:
	socket_tls_bad_cert(long code):
		socket_tls_error("Server certificate not verified: " + err_code(code)) { }
private:
	std::string err_code(long code);
};

class socket_tls_start_started: public socket_tls_error {
public:
	socket_tls_start_started():
		socket_tls_error("Starting TLS when already started", false) { }
};

class socket_tls_unable_to_set_sni_host: public socket_tls_error {
public:
	socket_tls_unable_to_set_sni_host(std::string host):
		socket_tls_error("Set " + host + " for SNI extension failed.") { }
};

class socket_tls_session_sharing_fail: public socket_tls_error {
public:
	socket_tls_session_sharing_fail(std::string msg):
		socket_tls_error("Session sharing failed: " + msg,
				 false) { }
	socket_tls_session_sharing_fail():
		socket_tls_error("Session sharing failed") { }
};

}

#endif
